# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import os
import copy
import json

from br4nch.utility.utility_librarian import UtilityLibrarian
from br4nch.utility.utility_handler import UtilityHandler


class ExportJson:
    def __init__(self, tree, output_folder, data_types=False):
        self.trees = tree
        self.output_folders = output_folder
        self.data_types = data_types

        self.nodes = []
        self.duplicates = []
        self.sorted_duplicates = {}
        self.content = []

        self.validate_arguments()
        self.task_manager()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.output_folders, list):
            self.output_folders = [self.output_folders]

        for folder in self.output_folders:
            if not isinstance(folder, str):
                raise UtilityHandler.InstanceStringError("output_folder", folder)

            if not os.path.isdir(folder):
                raise UtilityHandler.NotExistingDirectoryError(folder)

        if not self.data_types:
            if not isinstance(self.data_types, bool):
                raise UtilityHandler.InstanceBooleanError("data_types", self.data_types)

    def task_manager(self):
        for tree in self.trees:
            structure = copy.deepcopy(UtilityLibrarian.existing_trees[tree][list(UtilityLibrarian.existing_trees[tree])[0]])

            self.get_nodes(structure)
            self.sort_duplicates()
            self.update_node(structure)
            self.get_content(structure)

            while True:
                self.convert_to_dict(structure)
                if not self.content:
                    break

            for folder in self.output_folders:
                with open(folder + "/br4nch-" + tree + ".json", 'w', encoding='utf-8') as file:
                    json.dump(structure, file, indent=4)

    def get_nodes(self, nested_dictionary):
        for parent, children in nested_dictionary.copy().items():
            if children:
                self.nodes.append(parent[:-15])
                self.get_nodes(children)

    def sort_duplicates(self):
        for node in self.nodes:
            self.sorted_duplicates.update({node: []})

            for count in range(self.nodes.count(node)):
                self.sorted_duplicates[node].append(node + "#" + str(count + 1))

            if len(self.sorted_duplicates[node]) < 2:
                self.sorted_duplicates.pop(node)

    def update_node(self, nested_dictionary):
        for parent, children in nested_dictionary.copy().items():
            if parent[:-15] in self.sorted_duplicates and self.sorted_duplicates[parent[:-15]]:
                updated_parent = self.sorted_duplicates[parent[:-15]][0]
                self.sorted_duplicates[parent[:-15]].pop(0)
            else:
                updated_parent = parent[:-15]

            nested_dictionary.update({updated_parent: children})
            nested_dictionary.pop(parent)
            self.update_node(children)

    def get_content(self, nested_dictionary):
        for parent, children in nested_dictionary.items():
            if {parent: children} not in self.content:
                self.content.append({parent: children})

            self.get_content(children)

    def convert_to_dict(self, nested_dictionary):
        if isinstance(nested_dictionary, dict):
            for parent, children in nested_dictionary.items():
                if self.content:
                    if {parent: children} == self.content[-1]:
                        saved_children = []
                        set_list = False

                        self.content.pop(-1)

                        for grandchild, great_grandchildren in children.items():
                            saved_children.append({grandchild: great_grandchildren})

                        for index in range(len(saved_children)):
                            for grandchild, great_grandchildren in saved_children[index].items():
                                if isinstance(great_grandchildren, dict) and not great_grandchildren:
                                    set_list = True
                                    saved_children[index] = grandchild

                        if set_list:
                            for index in range(len(saved_children)):
                                if self.data_types:
                                    if saved_children[index] and isinstance(saved_children[index], str):
                                        if saved_children[index].lower() == "none" \
                                                or saved_children[index].lower() == "null":
                                            saved_children[index] = None
                                        elif saved_children[index].lower() == "true":
                                            saved_children[index] = True
                                        elif saved_children[index].lower() == "false":
                                            saved_children[index] = False
                                        elif saved_children[index].isdigit():
                                            saved_children[index] = int(saved_children[index])
                                        else:
                                            try:
                                                saved_children[index] = float(saved_children[index])
                                            except ValueError:
                                                pass

                            if len(saved_children) > 1:
                                nested_dictionary.update({parent: saved_children})
                            else:
                                if saved_children:
                                    nested_dictionary.update({parent: saved_children[0]})

                    self.convert_to_dict(children)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class CreateTree:
    def __init__(self, tree, header):
        self.trees = tree
        self.header = header

        self.validate_arguments()
        self.create_tree()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        for tree in self.trees:
            if not isinstance(tree, str):
                raise UtilityHandler.InstanceStringError("tree", tree)

            if not tree.isalnum():
                raise UtilityHandler.InvalidTreeNameError(tree)

            if tree.lower() in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.DuplicateTreeError(tree)

        if not isinstance(self.header, str):
            raise UtilityHandler.InstanceStringError("header", self.header)

    def create_tree(self):
        for tree in self.trees:
            UtilityLibrarian.existing_trees.update({tree: {self.header: {}}})
            UtilityLibrarian.existing_output.update({tree: []})
            UtilityLibrarian.existing_sizes.update({tree: 0})
            UtilityLibrarian.existing_symbols.update({tree: {"line": "┃", "split": "┣━", "end": "┗━"}})



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import copy

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class DuplicateTree:
    def __init__(self, new_tree, target_tree, attributes=False):
        self.new_tree = new_tree
        self.target_tree = target_tree
        self.attributes = attributes

        self.validate_arguments()
        self.duplicate_tree()

    def validate_arguments(self):
        if not isinstance(self.new_tree, str):
            raise UtilityHandler.InstanceStringError("new_tree", self.new_tree)

        if not self.new_tree.isalnum():
            raise UtilityHandler.InvalidTreeNameError(self.new_tree)

        if self.new_tree.lower() in list(map(str.lower, UtilityLibrarian.existing_trees)):
            raise UtilityHandler.DuplicateTreeError(self.new_tree)

        if not isinstance(self.target_tree, str):
            raise UtilityHandler.InstanceStringError("target_tree", self.target_tree)

        if self.target_tree.lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
            raise UtilityHandler.NotExistingTreeError(self.target_tree)

        for existing_tree in list(UtilityLibrarian.existing_trees):
            if self.target_tree.lower() == existing_tree.lower():
                self.target_tree = existing_tree

        if self.attributes:
            if not isinstance(self.attributes, bool):
                raise UtilityHandler.InstanceBooleanError("attributes", self.attributes)

    def duplicate_tree(self):
        UtilityLibrarian.existing_trees.update(
            {self.new_tree: copy.deepcopy(UtilityLibrarian.existing_trees[self.target_tree])})
        UtilityLibrarian.existing_output.update({self.new_tree: []})

        if self.attributes:
            UtilityLibrarian.existing_sizes.update({self.new_tree: UtilityLibrarian.existing_sizes[self.target_tree]})
            UtilityLibrarian.existing_symbols.update(
                {self.new_tree: UtilityLibrarian.existing_symbols[self.target_tree]})
        else:
            UtilityLibrarian.existing_sizes.update({self.new_tree: 0})
            UtilityLibrarian.existing_symbols.update({self.new_tree: {"line": "┃", "split": "┣━", "end": "┗━"}})



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from br4nch.utility.utility_librarian import UtilityLibrarian
from br4nch.utility.utility_handler import UtilityHandler
from br4nch.utility.utility_generator import UtilityGenerator
from br4nch.display.display_tree import DisplayTree


class GetNode:
    def __init__(self, tree, node="", sensitive=False, include="", exclude="", beautify=True):
        self.trees = tree
        self.nodes = node
        self.sensitive = sensitive
        self.includes = include
        self.excludes = exclude
        self.beautify = beautify

        self.validate_arguments()
        self.manage_package()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if self.nodes:
            if not isinstance(self.nodes, list):
                self.nodes = [self.nodes]

            for node in self.nodes:
                if not isinstance(node, str):
                    raise UtilityHandler.InstanceStringError("node", node)

        if self.sensitive:
            if not isinstance(self.sensitive, bool):
                raise UtilityHandler.InstanceBooleanError("sensitive", self.sensitive)

        if self.includes:
            if not isinstance(self.includes, list):
                self.includes = [self.includes]

            for include in self.includes:
                if not isinstance(include, str):
                    raise UtilityHandler.InstanceStringError("include", include)

        if self.excludes:
            if not isinstance(self.excludes, list):
                self.excludes = [self.excludes]

            for exclude in self.excludes:
                if not isinstance(exclude, str):
                    raise UtilityHandler.InstanceStringError("exclude", exclude)

        if self.beautify:
            if not isinstance(self.beautify, bool):
                raise UtilityHandler.InstanceBooleanError("beautify", self.beautify)

    def manage_package(self):
        tree_package = []

        for tree in self.trees:
            levels = [0]
            self.elevator(levels, UtilityLibrarian.existing_trees[tree][list(UtilityLibrarian.existing_trees[tree])[0]])
            levels.append(0)

            if self.nodes:
                for node in self.nodes:
                    tree_package = self.get_node(tree, node, levels, [0],
                                                 UtilityLibrarian.existing_trees[tree][list(
                                                     UtilityLibrarian.existing_trees[tree])[0]], tree_package, "")
            else:
                tree_package = self.get_all_nodes(tree, levels, [0],
                                                  UtilityLibrarian.existing_trees[tree][list(
                                                      UtilityLibrarian.existing_trees[tree])[0]], tree_package, "")

        if tree_package and self.beautify:
            tree_uid = UtilityGenerator(True).generate_uid()

            UtilityLibrarian.existing_trees.update({tree_uid: {"Get Node Result:": {}}})
            UtilityLibrarian.existing_output.update({tree_uid: []})
            UtilityLibrarian.existing_sizes.update({tree_uid: 0})
            UtilityLibrarian.existing_symbols.update({tree_uid: {"line": "┃", "split": "┣━", "end": "┗━"}})

            for box in tree_package:
                if box[0] not in UtilityLibrarian.existing_trees[tree_uid][list(
                        UtilityLibrarian.existing_trees[tree_uid])[0]]:
                    UtilityLibrarian.existing_trees[tree_uid][list(
                        UtilityLibrarian.existing_trees[tree_uid])[0]].update({box[0]: {}})

                if box[1] not in UtilityLibrarian.existing_trees[tree_uid][list(
                        UtilityLibrarian.existing_trees[tree_uid])[0]][box[0]]:
                    UtilityLibrarian.existing_trees[tree_uid][list(
                        UtilityLibrarian.existing_trees[tree_uid])[0]][box[0]].update({box[1]: {}})

                if box[2] not in UtilityLibrarian.existing_trees[tree_uid][list(
                        UtilityLibrarian.existing_trees[tree_uid])[0]][box[0]][box[1]]:
                    UtilityLibrarian.existing_trees[tree_uid][list(
                        UtilityLibrarian.existing_trees[tree_uid])[0]][box[0]][box[1]].update({box[2]: {}})

            self.update_tree(tree_uid, UtilityLibrarian.existing_trees[tree_uid])

            DisplayTree(tree_uid, True)

    def elevator(self, levels, nested_dictionary, height=0):
        for children in nested_dictionary.values():
            levels.append(height)
            self.elevator(levels, children, height + 1)

    def get_node(self, tree, node, levels, trace, nested_dictionary, tree_package, visual_position):
        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1
            trace[0] = trace[0] + 1
            skip = False

            if levels[trace[0]] <= levels[trace[0] - 1]:
                visual_position = visual_position[:-2]
            visual_position = visual_position + "." + str(count)

            if self.includes:
                for include in self.includes:
                    if include not in parent[:-15]:
                        skip = True

            if self.excludes:
                for exclude in self.excludes:
                    if exclude in parent[:-15]:
                        skip = True

            if not skip:
                if self.sensitive:
                    if parent[:-15] == node:
                        for character in visual_position:
                            if character == ".":
                                visual_position = visual_position[1:]
                            else:
                                break

                        tree_package.append([tree, parent[:-15], visual_position])
                        if not self.beautify:
                            print(visual_position)
                else:
                    if parent[:-15].lower() == node.lower():
                        for character in visual_position:
                            if character == ".":
                                visual_position = visual_position[1:]
                            else:
                                break

                        tree_package.append([tree, parent[:-15], visual_position])
                        if not self.beautify:
                            print(visual_position)

            if children:
                self.get_node(tree, node, levels, trace, children, tree_package, visual_position)

        return tree_package

    def get_all_nodes(self, tree, levels, trace, nested_dictionary, tree_package, visual_position):
        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1
            trace[0] = trace[0] + 1
            skip = False

            if levels[trace[0]] <= levels[trace[0] - 1]:
                visual_position = visual_position[:-2]
            visual_position = visual_position + "." + str(count)

            for character in visual_position:
                if character == ".":
                    visual_position = visual_position[1:]
                else:
                    break

            if self.includes:
                for include in self.includes:
                    if include not in parent[:-15]:
                        skip = True

            if self.excludes:
                for exclude in self.excludes:
                    if exclude in parent[:-15]:
                        skip = True

            if not skip:
                tree_package.append([tree, parent[:-15], visual_position])
                if not self.beautify:
                    print(visual_position)

            if children:
                self.get_all_nodes(tree, levels, trace, children, tree_package, visual_position)

        return tree_package

    def update_tree(self, tree, nested_dictionary, height=0):
        for parent, children in nested_dictionary.copy().items():
            if height == 1:
                nested_dictionary["Tree: " + parent + UtilityGenerator().generate_uid()] = \
                    nested_dictionary.pop(parent)

            if height == 2:
                nested_dictionary["Node: " + parent + UtilityGenerator().generate_uid()] = \
                    nested_dictionary.pop(parent)

            if height == 3:
                nested_dictionary["Position: " + parent + UtilityGenerator().generate_uid()] = \
                    nested_dictionary.pop(parent)

            if children:
                self.update_tree(tree, children, height + 1)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import os
import json

from br4nch.utility.utility_librarian import UtilityLibrarian
from br4nch.utility.utility_handler import UtilityHandler
from br4nch.utility.utility_generator import UtilityGenerator


class LoadJson:
    def __init__(self, new_tree, header, json_file):
        self.new_trees = new_tree
        self.header = header
        self.json_file = json_file

        self.existing = []

        self.validate_arguments()
        self.task_manager()

    def validate_arguments(self):
        if not isinstance(self.new_trees, list):
            self.new_trees = [self.new_trees]

        for tree in self.new_trees:
            if not isinstance(tree, str):
                raise UtilityHandler.InstanceStringError("new_trees", tree)

            if not tree.isalnum():
                raise UtilityHandler.InvalidTreeNameError(tree)

            if tree.lower() in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.DuplicateTreeError(tree)

        if not isinstance(self.header, str):
            raise UtilityHandler.InstanceStringError("header", self.header)

        if not isinstance(self.json_file, str):
            raise UtilityHandler.InstanceStringError("json_file", self.json_file)

        if not os.path.isfile(self.json_file):
            raise UtilityHandler.NotExistingJsonFileError(self.json_file)

    def task_manager(self):
        file = open(self.json_file)
        json_content = json.loads(json.dumps(json.load(file)))
        file.close()

        while True:
            self.convert_non_strings(json_content)
            converted = self.convert_lists(self.get_lists(json_content))

            if not converted:
                break

        self.add_uid(json_content)

        for tree in self.new_trees:
            UtilityLibrarian.existing_output.update({tree: []})
            UtilityLibrarian.existing_sizes.update({tree: 0})
            UtilityLibrarian.existing_symbols.update({tree: {"line": "┃", "split": "┣━", "end": "┗━"}})
            UtilityLibrarian.existing_trees.update({tree: {self.header: json_content}})

    def convert_non_strings(self, nested_dictionary):
        if not isinstance(nested_dictionary, dict):
            nested_dictionary = {str(nested_dictionary): {}}

        for parent, children in nested_dictionary.items():
            if children is None:
                children = "null"
            if isinstance(children, bool) and children is True:
                children = "true"
            if isinstance(children, bool) and children is False:
                children = "false"

            if not isinstance(children, dict) and not isinstance(children, list):
                if children not in self.existing:
                    children = str(children) + UtilityGenerator().generate_uid()
                    self.existing.append(children)

                nested_dictionary.update({parent: {str(children): {}}})

            self.convert_non_strings(children)

    def get_lists(self, nested_dictionary):
        for parent, children in nested_dictionary.items():
            if isinstance(children, list):
                return nested_dictionary
            else:
                returned_dictionary = self.get_lists(children)
                if returned_dictionary:
                    return returned_dictionary

    def convert_lists(self, nested_dictionary):
        if not nested_dictionary:
            return

        converted = False
        for parent, children in nested_dictionary.items():
            saved_children = []

            if isinstance(children, dict):
                for grandchild, great_grandchildren in children.items():
                    saved_children.append({grandchild: great_grandchildren})
            else:
                for child in children:
                    saved_children.append(child)

            nested_dictionary.update({parent: {}})

            if saved_children:
                converted = True
                for saved_child in saved_children:
                    if not isinstance(saved_child, dict) and not isinstance(saved_child, list):
                        saved_child = {str(saved_child): {}}

                    for grandchild, great_grandchildren in saved_child.items():
                        if not isinstance(great_grandchildren, dict) and not isinstance(great_grandchildren, list):
                            great_grandchildren = {str(great_grandchildren): {}}

                        if grandchild not in self.existing:
                            grandchild = str(grandchild) + UtilityGenerator().generate_uid()
                            self.existing.append(grandchild)

                        nested_dictionary[parent].update({str(grandchild): great_grandchildren})

        return converted

    def add_uid(self, nested_dictionary):
        for parent, children in nested_dictionary.copy().items():
            if not isinstance(parent, dict) and not isinstance(parent, list):
                nested_dictionary.pop(parent)

                if parent not in self.existing:
                    parent = str(parent) + UtilityGenerator().generate_uid()
                    self.existing.append(parent)

                nested_dictionary.update({parent: children})

            self.add_uid(children)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import copy

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_generator import UtilityGenerator
from ..utility.utility_decider import UtilityDecider


class DuplicateNode:
    def __init__(self, tree, duplicate_node, to_parent="", target_tree="", delete=False):
        self.trees = tree
        self.duplicate_nodes = duplicate_node
        self.to_parents = to_parent
        self.target_trees = target_tree
        self.delete = delete

        self.validate_arguments()
        self.duplicate_node()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.duplicate_nodes, list):
            self.duplicate_nodes = [self.duplicate_nodes]

        if not isinstance(self.to_parents, list):
            self.to_parents = [self.to_parents]

        if self.target_trees:
            if not isinstance(self.target_trees, list):
                self.target_trees = [self.target_trees]

            if "*" in self.target_trees:
                self.target_trees.clear()
                for existing_tree in list(UtilityLibrarian.existing_trees):
                    self.target_trees.append(existing_tree)

            for index in range(len(self.target_trees)):
                if not isinstance(self.target_trees[index], str):
                    raise UtilityHandler.InstanceStringError("target_tree", self.target_trees[index])

                if self.target_trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                    raise UtilityHandler.NotExistingTreeError(self.target_trees[index])

                for existing_tree in list(UtilityLibrarian.existing_trees):
                    if self.target_trees[index].lower() == existing_tree.lower():
                        self.target_trees[index] = existing_tree

        if self.delete:
            if not isinstance(self.delete, bool):
                raise UtilityHandler.InstanceBooleanError("delete", self.delete)

    def duplicate_node(self):
        for tree in self.trees:
            queue_delete = []
            queue_add = []

            for node_position in UtilityDecider(tree, "to_parent",
                                                self.duplicate_nodes.copy()).get_formatted_positions():
                for parent_position in UtilityDecider(tree, "to_parent",
                                                      self.to_parents.copy()).get_formatted_positions():
                    children = self.get_nodes(tree, node_position, [],
                                              UtilityLibrarian.existing_trees[tree][list(
                                                  UtilityLibrarian.existing_trees[tree])[0]])

                    if children:
                        queue_delete.append(children[1])

                        queue_add.append([children[0], self.get_nodes(
                            tree, [], parent_position, UtilityLibrarian.existing_trees[tree][list(
                                UtilityLibrarian.existing_trees[tree])[0]])])

                        if self.target_trees:
                            for sibling in self.target_trees:
                                queue_add.append([children[0], self.get_nodes(
                                    sibling, [], parent_position,
                                    UtilityLibrarian.existing_trees[sibling][list(
                                        UtilityLibrarian.existing_trees[sibling])[0]])])

            if self.delete:
                for delete_node in queue_delete:
                    if delete_node:
                        for parent_node, child_nodes in delete_node.items():
                            del child_nodes[parent_node]

            for add_node in queue_add:
                if add_node and add_node[0] and add_node[1]:
                    self.change_node_uid(list(add_node[1])[0], add_node[0])
                    add_node[1][list(add_node[1])[0]].update(add_node[0])

    def get_nodes(self, tree, node_position, parent_position, nested_dictionary):
        if parent_position and parent_position[0] == "0":
            return {tree: UtilityLibrarian.existing_trees[tree][list(UtilityLibrarian.existing_trees[tree])[0]]}

        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1

            if node_position and count == int(node_position[0]):
                if len(node_position) == 1:
                    return [{parent: copy.deepcopy(children)}, {parent: nested_dictionary}]
                else:
                    if children:
                        node_position.pop(0)
                        return self.get_nodes(tree, node_position, parent_position, children)

            if parent_position and count == int(parent_position[0]):
                if len(parent_position) == 1:
                    return {tree: children}
                else:
                    if children:
                        parent_position.pop(0)
                        return self.get_nodes(tree, node_position, parent_position, children)

    def change_node_uid(self, tree, nested_dictionary):
        for parent, children in nested_dictionary.copy().items():
            parent_node_uid = parent[:-15] + UtilityGenerator().generate_uid()

            nested_dictionary[parent_node_uid] = nested_dictionary.pop(parent)

            if children:
                self.change_node_uid(tree, children)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import uuid

from ..utility.utility_librarian import UtilityLibrarian


class UtilityGenerator:
    def __init__(self, tree=False):
        self.tree = tree

        self.generate_uid()

    def generate_uid(self):
        uid = str(uuid.uuid4()).replace("-", "")[0:10]

        if self.tree == "-":
            if uid not in list(UtilityLibrarian.existing_trees):
                return uid
        else:
            return ":uid=" + uid



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

class UtilityHandler:
    class InstanceStringError(Exception):
        __module__ = Exception.__module__

        def __init__(self, argument, value):
            self.argument = argument
            self.value = value

        def __str__(self):
            return "The " + str(self.argument) + " argument: '" + str(self.value) \
                   + "' must be an instance of a 'str' and not '" + str(type(self.value).__name__) + "'."

    class InstanceBooleanError(Exception):
        __module__ = Exception.__module__

        def __init__(self, argument, value):
            self.argument = argument
            self.value = value

        def __str__(self):
            return "The " + str(self.argument) + " argument: '" + str(self.value) \
                   + "' must be an instance of a 'bool' and not '" + str(type(self.value).__name__) + "'."

    class InstanceIntegerError(Exception):
        __module__ = Exception.__module__

        def __init__(self, argument, value):
            self.argument = argument
            self.value = value

        def __str__(self):
            return "The " + str(self.argument) + " argument: '" + str(self.value) \
                   + "' must be an instance of a 'int' and not '" + str(type(self.value).__name__) + "'."

    class InvalidTreeNameError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The tree name: '" + self.value \
                   + "' is not valid. Only numbers and/or letters may be used to add a tree."

    class InvalidPositionError(Exception):
        __module__ = Exception.__module__

        def __init__(self, argument, value):
            self.argument = argument
            self.value = value

        def __str__(self):
            return "The " + self.argument + ": '" + self.value \
                   + "' is not valid. Only valid positions, operators and existing nodes can be used to decide the " \
                     "position for the " + self.argument + " argument."

    class InvalidSizeError(Exception):
        __module__ = Exception.__module__

        def __str__(self):
            return "The sizes that can be used is '0-20'."

    class InvalidTreeFileError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The file: '" + self.value + "' is not valid as a tree file."

    class InvalidAttributesFileError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The file: '" + self.value + "' is not valid as a attributes file."

    class NotExistingTreeError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The tree: '" + str(self.value) + "' does not exists."

    class NotExistingDirectoryError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The directory: '" + str(self.value) + "' does not exist."

    class NotExistingTreeFileError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The tree file: '" + str(self.value) + "' does not exist."

    class NotExistingAttributesFileError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The attributes file: '" + str(self.value) + "' does not exist."

    class NotExistingJsonFileError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The json file: '" + str(self.value) + "' does not exist."

    class NotSizeableError(Exception):
        __module__ = Exception.__module__

        def __str__(self):
            return "You can only use the 'size' function with the default symbols."

    class DuplicateTreeError(Exception):
        __module__ = Exception.__module__

        def __init__(self, value):
            self.value = value

        def __str__(self):
            return "The tree: '" + str(self.value) + "' already exists."

    class RequiredSymbolChangeError(Exception):
        __module__ = Exception.__module__

        def __str__(self):
            return "Change at least one of the given symbols: 'line', 'split' or 'end'."



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import os

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_generator import UtilityGenerator


class LoadFolder:
    def __init__(self, new_tree, folder_path, header="", include="", exclude="", unused=True, folder_priority=True):
        self.new_trees = new_tree
        self.folder_path = folder_path
        self.header = header
        self.includes = include
        self.excludes = exclude
        self.unused = unused
        self.folder_priority = folder_priority

        self.validate_arguments()
        self.load_folder()

    def validate_arguments(self):
        if not isinstance(self.new_trees, list):
            self.new_trees = [self.new_trees]

        for tree in self.new_trees:
            if not isinstance(tree, str):
                raise UtilityHandler.InstanceStringError("new_tree", tree)

            if not tree.isalnum():
                raise UtilityHandler.InvalidTreeNameError(tree)

            if tree.lower() in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.DuplicateTreeError(tree)

        if not isinstance(self.folder_path, str):
            raise UtilityHandler.InstanceStringError("folder_path", self.folder_path)

        if not os.path.isdir(self.folder_path):
            raise UtilityHandler.NotExistingDirectoryError(self.folder_path)

        if self.header:
            if not isinstance(self.header, str):
                raise UtilityHandler.InstanceStringError("header", self.header)

        if not self.header:
            self.header = self.folder_path

        if self.includes:
            if not isinstance(self.includes, list):
                self.includes = [self.includes]

            for index in range(len(self.includes)):
                if not isinstance(self.includes[index], str):
                    raise UtilityHandler.InstanceStringError("include", self.includes[index])

                if not self.includes[index][0] == ".":
                    self.includes[index] = "." + self.includes[index]

        if self.excludes:
            if not isinstance(self.excludes, list):
                self.excludes = [self.excludes]

            for index in range(len(self.excludes)):
                if not isinstance(self.excludes[index][0], str):
                    raise UtilityHandler.InstanceStringError("exclude", self.excludes[index][0])

                if not self.excludes[index][0] == ".":
                    self.excludes[index] = "." + self.excludes[index]

        if self.unused:
            if not isinstance(self.unused, bool):
                raise UtilityHandler.InstanceBooleanError("unused", self.unused)

        if self.folder_priority:
            if not isinstance(self.folder_priority, bool):
                raise UtilityHandler.InstanceBooleanError("folder_priority", self.folder_priority)

            if self.folder_priority:
                self.folder_priority = False
            else:
                self.folder_priority = True

    def load_folder(self):
        for tree in self.new_trees:
            tree_structure = {self.header: {}}
            queue_delete = []
            paths = []
            roots = []

            for root, directories, files in os.walk(self.folder_path, topdown=self.folder_priority):
                paths.append(root.replace("\\", "/"))
                roots.append(root.replace("\\", "/"))

                for file in files:
                    paths.append(root.replace("\\", "/") + "/" + file)

            path_length = len(self.folder_path.replace("\\", "/").split("/"))

            for index in range(len(paths)):
                if self.excludes:
                    for exclude in self.excludes:
                        if paths[index][-len(exclude):] == exclude:
                            if paths[index] not in queue_delete:
                                queue_delete.append(paths[index])

                if self.includes:
                    for include in self.includes:
                        if paths[index][-len(include):] not in self.includes and paths[index] not in roots:
                            if paths[index] not in queue_delete:
                                queue_delete.append(paths[index])

            for path in queue_delete:
                paths.remove(path)

            if not self.unused:
                for root in roots:
                    hit = 0
                    for path in paths:
                        if root in path:
                            hit = hit + 1
                    if hit == 1:
                        if root in paths:
                            paths.remove(root)

            for path in paths:
                for number in range(len(path.split("/")[path_length:])):
                    previous_file = []

                    if number - 1 >= 0:
                        previous_file = path.split("/")[path_length:number - len(path.split("/")[path_length:])]

                    self.create_tree(
                        path.split("/")[path_length:][number], [self.header] + previous_file, tree_structure)

            UtilityLibrarian.existing_output.update({tree: []})
            UtilityLibrarian.existing_sizes.update({tree: 0})
            UtilityLibrarian.existing_symbols.update({tree: {"line": "┃", "split": "┣━", "end": "┗━"}})

            self.generate_nodes_uid(tree, tree_structure[list(tree_structure)[0]])

            UtilityLibrarian.existing_trees.update({tree: tree_structure})

    def create_tree(self, file, previous_file, child):
        for parent_node, child_nodes in child.items():
            if previous_file and parent_node == previous_file[0]:
                previous_file.pop(0)

                if not previous_file and file not in child_nodes:
                    return child_nodes.update({file: {}})
                else:
                    return self.create_tree(file, previous_file, child_nodes)

    def generate_nodes_uid(self, tree, child):
        for parent_node, child_nodes in child.copy().items():
            parent_node_uid = parent_node + UtilityGenerator().generate_uid()

            child[parent_node_uid] = child.pop(parent_node)

            if child_nodes:
                self.generate_nodes_uid(tree, child_nodes)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class DeleteTree:
    def __init__(self, tree):
        self.trees = tree

        self.validate_arguments()
        self.delete_tree()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for tree in self.trees:
            if not isinstance(tree, str):
                raise UtilityHandler.InstanceStringError("tree", tree)

            if tree.lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(tree)

    def delete_tree(self):
        for tree in self.trees:
            del UtilityLibrarian.existing_trees[tree]
            del UtilityLibrarian.existing_output[tree]
            del UtilityLibrarian.existing_sizes[tree]
            del UtilityLibrarian.existing_symbols[tree]



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import copy

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_generator import UtilityGenerator
from ..display.display_tree import DisplayTree


class DisplayAssist:
    def __init__(self, tree):
        self.trees = tree

        self.validate_arguments()
        self.display_assist()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

    def display_assist(self):
        for tree in self.trees:
            levels = [0]
            self.elevator(levels, UtilityLibrarian.existing_trees[tree][list(UtilityLibrarian.existing_trees[tree])[0]])
            levels.append(0)

            tree_uid = UtilityGenerator(True).generate_uid()

            UtilityLibrarian.existing_trees.update({tree_uid: copy.deepcopy(UtilityLibrarian.existing_trees[tree])})
            UtilityLibrarian.existing_trees[tree_uid][str("0: " + list(UtilityLibrarian.existing_trees[tree])[0])] = \
                UtilityLibrarian.existing_trees[tree_uid].pop(list(UtilityLibrarian.existing_trees[tree_uid])[0])
            UtilityLibrarian.existing_output.update({tree_uid: []})
            UtilityLibrarian.existing_sizes.update({tree_uid: 0})
            UtilityLibrarian.existing_symbols.update(
                {tree_uid: {"line": "┃", "split": "┣━", "end": "┗━"}})

            self.set_node_positions(levels, [0],
                                    UtilityLibrarian.existing_trees[tree_uid][list(
                                        UtilityLibrarian.existing_trees[tree_uid])[0]])

            DisplayTree(tree_uid, True)

    def elevator(self, levels, nested_dictionary, height=0):
        for children in nested_dictionary.values():
            levels.append(height)
            self.elevator(levels, children, height + 1)

    def set_node_positions(self, levels, trace, nested_dictionary, position=""):
        count = 0
        for parent, children in nested_dictionary.copy().items():
            count = count + 1
            trace[0] = trace[0] + 1

            if levels[trace[0]] <= levels[trace[0] - 1]:
                position = position[:-2]
            position = position + "." + str(count)

            nested_dictionary[position[1:] + ": " + parent] = nested_dictionary.pop(parent)

            if children:
                self.set_node_positions(levels, trace, children, position)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian


class UtilityBuilder:
    def __init__(self, tree):
        self.tree = tree

        self.manager()

    def manager(self):
        UtilityLibrarian.existing_output[self.tree].append(list(UtilityLibrarian.existing_trees[self.tree])[0])

        levels = [0]
        self.elevator(
            levels, UtilityLibrarian.existing_trees[self.tree][list(UtilityLibrarian.existing_trees[self.tree])[0]])
        levels.append(0)

        self.build_nodes(
            [levels, [0], [], [], [0]],
            UtilityLibrarian.existing_trees[self.tree][list(UtilityLibrarian.existing_trees[self.tree])[0]])

    def elevator(self, levels, nested_dictionary, height=0):
        for children in nested_dictionary.values():
            levels.append(height)
            self.elevator(levels, children, height + 1)

    def build_nodes(self, package, nested_dictionary):
        for parent, children in nested_dictionary.items():
            package[1][0] = package[1][0] + 1

            extend = self.build_extender(parent, nested_dictionary, package[0], package[1], package[2], package[3],
                                         package[4])

            size = ""
            for _ in range(UtilityLibrarian.existing_sizes[self.tree]):
                size = size + extend + UtilityLibrarian.existing_symbols[self.tree].get("line") + "\n"

            if parent != list(nested_dictionary)[-1]:
                UtilityLibrarian.existing_output[self.tree].append(
                    size + extend + UtilityLibrarian.existing_symbols[self.tree].get("split") + " "
                    + parent[:-15].replace("\n", "\n" + extend
                                           + UtilityLibrarian.existing_symbols[self.tree].get("line") + " "
                                           * int(len(UtilityLibrarian.existing_symbols[self.tree].get("split")))))
            else:
                UtilityLibrarian.existing_output[self.tree].append(
                    size + extend + UtilityLibrarian.existing_symbols[self.tree].get("end") + " "
                    + parent[:-15].replace("\n", "\n" + extend + " " * int(
                        len(UtilityLibrarian.existing_symbols[self.tree].get("end")) + 1)))

            if children:
                self.build_nodes(package, children)

    def build_extender(self, node, child, levels, trace, chain, queue, last):
        if levels[trace[0]] != 0:
            if levels[trace[0]] != levels[trace[0] - 1]:
                if queue or last[0] == len(
                        list(UtilityLibrarian.existing_trees[self.tree][list(
                            UtilityLibrarian.existing_trees[self.tree])[0]])):
                    chain.append(" " * int(len(UtilityLibrarian.existing_symbols[self.tree].get("end")) + 1))
                    queue.clear()

                    if last[0] == len(
                            list(UtilityLibrarian.existing_trees[self.tree][list(
                                UtilityLibrarian.existing_trees[self.tree])[0]])):
                        last[0] = 0
                else:
                    chain.append(str(UtilityLibrarian.existing_symbols[self.tree].get("line") + " "
                                     * int(len(UtilityLibrarian.existing_symbols[self.tree].get("split")))))

                if levels[trace[0]] < levels[trace[0] - 1]:
                    for _ in range(-1, int(levels[trace[0] - 1] - levels[trace[0]])):
                        chain.pop()

            if node == list(child)[-1]:
                if levels[trace[0]] < levels[trace[0] + 1]:
                    queue.append(levels[trace[0]])
        else:
            chain.clear()
            last[0] = last[0] + 1

        extend = ""
        for line in chain:
            extend = extend + line
        return extend



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import os

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_builder import UtilityBuilder


class ExportText:
    def __init__(self, tree, output_folder):
        self.trees = tree
        self.output_folders = output_folder

        self.validate_arguments()
        self.export_text()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.output_folders, list):
            self.output_folders = [self.output_folders]

        for folder in self.output_folders:
            if not isinstance(folder, str):
                raise UtilityHandler.InstanceStringError("output_folder", folder)

            if not os.path.isdir(folder):
                raise UtilityHandler.NotExistingDirectoryError(folder)

    def export_text(self):
        for tree in self.trees:
            for folder in self.output_folders:
                with open(folder + "/br4nch-" + tree + ".txt", 'w', encoding='utf-8') as file:
                    UtilityBuilder(tree)

                    for line in UtilityLibrarian.existing_output[tree]:
                        file.write(line + "\n")
                    UtilityLibrarian.existing_output[tree].clear()



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from .get_tree import GetTree as Tree
from .get_node import GetNode as Node
from .get_position import GetPosition as Position



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_generator import UtilityGenerator
from ..utility.utility_decider import UtilityDecider


class ReplaceNode:
    def __init__(self, tree, new_node, target_node):
        self.trees = tree
        self.new_node = new_node
        self.target_nodes = target_node

        self.validate_arguments()
        self.replace_node()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.new_node, str):
            raise UtilityHandler.InstanceStringError("new_node", self.new_node)

        while True:
            if self.new_node and self.new_node[-1] == "\n":
                self.new_node = self.new_node[:-1]
            else:
                break

        if not isinstance(self.target_nodes, list):
            self.target_nodes = [self.target_nodes]

    def replace_node(self):
        for tree in self.trees:
            for position in UtilityDecider(tree, "node", self.target_nodes.copy()).get_formatted_positions():
                child = self.get_nodes(position,
                                       UtilityLibrarian.existing_trees[tree][list(
                                           UtilityLibrarian.existing_trees[tree])[0]])
                if child:
                    for parent_node, child_nodes in child.items():
                        parent_node_uid = self.new_node + UtilityGenerator().generate_uid()

                        index = list(child_nodes).index(parent_node)
                        child_nodes[parent_node_uid] = child_nodes.pop(parent_node)

                        for index in list(child_nodes)[index:-1]:
                            child_nodes[index] = child_nodes.pop(index)

    def get_nodes(self, position, nested_dictionary):
        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    return {parent: nested_dictionary}
                else:
                    if children:
                        position.pop(0)
                        return self.get_nodes(position, children)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class UtilityDecider:
    def __init__(self, tree, argument, nodes):
        self.tree = tree
        self.argument = argument
        self.nodes = nodes

        self.nodes = FormatNode(self.tree, self.nodes).get_positions()
        self.positions = FormatPosition(self.tree, self.argument, self.nodes).get_position_package()

    def get_formatted_positions(self):
        for package in self.positions:
            while "" in package:
                package.remove("")

        return self.positions


class FormatNode:
    def __init__(self, tree, nodes):
        self.tree = tree
        self.nodes = nodes

        self.select = 0
        self.position = False

        self.manage_nodes()

    def manage_nodes(self):
        levels = [0]
        self.elevator(levels,
                      UtilityLibrarian.existing_trees[self.tree][list(
                          UtilityLibrarian.existing_trees[self.tree])[0]])
        levels.append(0)

        for node in self.nodes:
            self.format_node(node, levels, [0],
                             UtilityLibrarian.existing_trees[self.tree][list(
                                 UtilityLibrarian.existing_trees[self.tree])[0]], "")

        for node in self.nodes.copy():
            if "#" in node:
                skip = False
                for character in node.split("#")[-1]:
                    if not character.isnumeric():
                        skip = True

                if not skip:
                    split_node = node.rsplit("#", 1)
                    while "" in split_node:
                        split_node.remove("")

                    stripped_node = ""
                    if len(split_node) > 1:
                        for part in node.rsplit("#", 1)[:1]:
                            stripped_node = stripped_node + part
                        self.format_node_number(stripped_node, levels, [0],
                                                UtilityLibrarian.existing_trees[self.tree][list(
                                                    UtilityLibrarian.existing_trees[self.tree])[0]], "",
                                                int(node.split("#")[-1]), node)
            self.select = 0

    def elevator(self, levels, nested_dictionary, height=0):
        for children in nested_dictionary.values():
            levels.append(height)
            self.elevator(levels, children, height + 1)

    def format_node(self, node, levels, trace, nested_dictionary, visual_position):
        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1
            trace[0] = trace[0] + 1

            if levels[trace[0]] <= levels[trace[0] - 1]:
                visual_position = visual_position[:-2]
            visual_position = visual_position + "." + str(count)

            if parent[:-15].lower() == node.lower():
                for character in visual_position:
                    if character == ".":
                        visual_position = visual_position[1:]
                    else:
                        break

                stripped_node = ""
                for character in node.split("."):
                    stripped_node = stripped_node + character

                skip = False
                for character in stripped_node:
                    if not character.isnumeric():
                        skip = True

                if not skip:
                    self.validate_position(node.split("."),
                                           UtilityLibrarian.existing_trees[self.tree][list(
                                               UtilityLibrarian.existing_trees[self.tree])[0]])

                if skip or not self.position:
                    self.position = False
                    self.nodes.append(visual_position)

                    if node in self.nodes:
                        self.nodes.remove(node)

            if children:
                self.format_node(node, levels, trace, children, visual_position)

    def validate_position(self, position, nested_dictionary):
        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    self.position = True
                else:
                    if children:
                        position.pop(0)
                        return self.validate_position(position, children)

    def get_positions(self):
        return self.nodes

    def format_node_number(self, node, levels, trace, nested_dictionary, visual_position, number=0, full_node=""):
        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1
            trace[0] = trace[0] + 1

            if levels[trace[0]] <= levels[trace[0] - 1]:
                visual_position = visual_position[:-2]
            visual_position = visual_position + "." + str(count)

            if parent[:-15].lower() == node.lower():
                if full_node:
                    self.select = self.select + 1

                if not number or number == self.select:
                    for character in visual_position:
                        if character == ".":
                            visual_position = visual_position[1:]
                        else:
                            break

                    self.nodes.append(visual_position)
                    if full_node in self.nodes:
                        self.nodes.remove(full_node)

            if children:
                self.format_node_number(node, levels, trace, children, visual_position, number, full_node)


class FormatPosition:
    def __init__(self, tree, argument, position_package):
        self.tree = tree
        self.argument = argument
        self.position_package = position_package

        self.format_position(self.position_package)

    def format_position(self, position_package):
        for number in range(len(position_package)):
            if isinstance(position_package[number], list):
                for position in position_package[number]:
                    if not isinstance(position, str):
                        raise UtilityHandler.InstanceStringError(self.argument, position)
                    else:
                        for character in position:
                            if character not in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", ".", "/", "*", ">",
                                                 "<"]:
                                raise UtilityHandler.InvalidPositionError(self.argument, position_package[number])
            else:
                if not isinstance(position_package[number], str):
                    raise UtilityHandler.InstanceStringError("pos", position_package[number])
                else:
                    for character in position_package[number]:
                        if character not in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", ".", "/", "*", ">", "<"]:
                            raise UtilityHandler.InvalidPositionError(self.argument, position_package[number])

            if "." in position_package[number]:
                position_package[number] = position_package[number].split(".")
            else:
                if not isinstance(position_package[number], list):
                    position_package[number] = [position_package[number]]

        for number in range(len(position_package)):
            for position in range(len(position_package[number])):
                if "/" in position_package[number][position]:
                    for multiple_position in position_package[number][position].split("/"):
                        if not multiple_position:
                            raise UtilityHandler.InvalidPositionError(self.argument, position_package[number][position])

                    for count in range(len(position_package[number][position].split("/"))):
                        position_package.append(position_package[number].copy())
                        position_package[-1][position] = position_package[number][position].split("/")[count]
                    position_package.pop(number)

                    self.format_position(position_package)

        for number in range(len(position_package)):
            for position in range(len(position_package[number])):
                if "*" in position_package[number][position]:
                    if not UtilityLibrarian.existing_trees[self.tree][list(
                            UtilityLibrarian.existing_trees[self.tree])[0]]:
                        position_package[number].pop(position)
                        position_package[number].append("0")
                    else:
                        for count in range(self.calculate_operator(
                                position_package[number].copy(),
                                UtilityLibrarian.existing_trees[self.tree][list(
                                    UtilityLibrarian.existing_trees[self.tree])[0]])):
                            position_package.append(position_package[number].copy())
                            position_package[-1][position] = str(count + 1)
                        position_package.pop(number)

                if ">" in position_package[number][position]:
                    if ">" in position_package[number][position]:
                        for including_position in position_package[number][position].split(">"):
                            if not including_position:
                                raise UtilityHandler.InvalidPositionError(
                                    self.argument, position_package[number][position])

                    including_positions = position_package[number][position].split(">")
                    total_including_positions = len(including_positions)

                    for count in range(int(min(including_positions)), int(max(including_positions)) + 1):
                        including_positions.append(str(count))

                    for _ in range(total_including_positions):
                        including_positions.pop(0)

                    for count in range(len(including_positions)):
                        position_package.append(position_package[number].copy())
                        position_package[-1][position] = including_positions[count]
                    position_package.pop(number)

                if "<" in position_package[number][position]:
                    if "<" in position_package[number][position]:
                        for excluding_position in position_package[number][position].split("<"):
                            if not excluding_position:
                                raise UtilityHandler.InvalidPositionError(
                                    self.argument, position_package[number][position])

                    excluding_positions = []

                    for count in range(self.calculate_operator(
                            position_package[number].copy(),
                            UtilityLibrarian.existing_trees[self.tree][list(
                                UtilityLibrarian.existing_trees[self.tree])[0]])):
                        excluding_positions.append(str(count + 1))

                    for count in range(int(min(position_package[number][position].split("<"))),
                                       int(max(position_package[number][position].split("<"))) + 1):
                        excluding_positions.remove(str(count))

                    for count in range(len(excluding_positions)):
                        position_package.append(position_package[number].copy())
                        position_package[-1][position] = excluding_positions[count]
                    position_package.pop(number)

                if "/" in position_package[number][position] or "*" in position_package[number][position] \
                        or ">" in position_package[number][position] or "<" in position_package[number][position]:
                    self.format_position(position_package)

        return position_package

    def calculate_operator(self, position, nested_dictionary):
        count = 0
        for children in nested_dictionary.values():
            count = count + 1

            if "*" in position[0] or "<" in position[0] or count == int(position[0]):
                return len(nested_dictionary)

            if children and count == int(position[0]):
                position.pop(0)
                return self.calculate_operator(position, children)

    def get_position_package(self):
        return self.position_package



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class ReplaceHeader:
    def __init__(self, tree, new_header):
        self.trees = tree
        self.new_header = new_header

        self.validate_arguments()
        self.replace_header()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.new_header, str):
            raise UtilityHandler.InstanceStringError("new_header", self.new_header)

    def replace_header(self):
        for tree in self.trees:
            UtilityLibrarian.existing_trees[tree][self.new_header] = UtilityLibrarian.existing_trees[tree].pop(list(
                UtilityLibrarian.existing_trees[tree])[0])



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import copy

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_decider import UtilityDecider


class MoveNode:
    def __init__(self, tree, move_node, to_parent="", target_tree=""):
        self.trees = tree
        self.move_nodes = move_node
        self.to_parent = to_parent
        self.target_tree = target_tree

        self.validate_arguments()
        self.move_node()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.move_nodes, list):
            self.move_nodes = [self.move_nodes]

        if not self.to_parent:
            self.to_parent = "0"

        if not isinstance(self.to_parent, str):
            raise UtilityHandler.InstanceStringError("to_parent", self.to_parent)

        for parent in self.to_parent.split("."):
            if not parent.isnumeric():
                raise UtilityHandler.InvalidPositionError("to_parent", self.to_parent)

        if self.target_tree:
            if not isinstance(self.target_tree, str):
                raise UtilityHandler.InstanceStringError("target_tree", self.target_tree)

            if self.target_tree.lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.target_tree)

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.target_tree.lower() == existing_tree.lower():
                    self.target_tree = existing_tree

    def move_node(self):
        for tree in self.trees:
            queue_delete = []
            queue_add = []

            for node_position in UtilityDecider(tree, "move_node", self.move_nodes.copy()).get_formatted_positions():
                children = self.get_nodes(tree, node_position, [],
                                          UtilityLibrarian.existing_trees[tree][list(
                                              UtilityLibrarian.existing_trees[tree])[0]])

                if children:
                    queue_delete.append(children[1])

                    if self.target_tree:
                        tree = self.target_tree

                    queue_add.append([children[0], self.get_nodes(tree, [], self.to_parent.split("."),
                                                                  UtilityLibrarian.existing_trees[tree][list(
                                                                      UtilityLibrarian.existing_trees[tree])[0]])])

            for delete_node in queue_delete:
                if delete_node:
                    for parent_node, child_nodes in delete_node.items():
                        del child_nodes[parent_node]

            for add_node in queue_add:
                if add_node and add_node[0] and add_node[1]:
                    add_node[1][list(add_node[1])[0]].update(add_node[0])

    def get_nodes(self, tree, node_position, parent_position, nested_dictionary):
        if parent_position and parent_position[0] == "0":
            return {tree: UtilityLibrarian.existing_trees[tree][list(UtilityLibrarian.existing_trees[tree])[0]]}

        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1

            if node_position and count == int(node_position[0]):
                if len(node_position) == 1:
                    return [{parent: copy.deepcopy(children)}, {parent: nested_dictionary}]
                else:
                    if children:
                        node_position.pop(0)
                        return self.get_nodes(tree, node_position, parent_position, children)

            if parent_position and count == int(parent_position[0]):
                if len(parent_position) == 1:
                    return {tree: children}
                else:
                    if children:
                        parent_position.pop(0)
                        return self.get_nodes(tree, node_position, parent_position, children)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class ReplaceTree:
    def __init__(self, new_tree, target_tree):
        self.new_tree = new_tree
        self.target_tree = target_tree

        self.validate_arguments()
        self.replace_tree()

    def validate_arguments(self):
        if not isinstance(self.new_tree, str):
            raise UtilityHandler.InstanceStringError("new_tree", self.new_tree)

        if not self.new_tree.isalnum():
            raise UtilityHandler.InvalidTreeNameError(self.new_tree)

        if self.new_tree.lower() in list(map(str.lower, UtilityLibrarian.existing_trees)):
            raise UtilityHandler.DuplicateTreeError(self.new_tree)

        if not isinstance(self.target_tree, str):
            raise UtilityHandler.InstanceStringError("target_tree", self.target_tree)

        if self.target_tree.lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
            raise UtilityHandler.NotExistingTreeError(self.target_tree)

        for existing_tree in list(UtilityLibrarian.existing_trees):
            if self.target_tree.lower() == existing_tree.lower():
                self.target_tree = existing_tree

    def replace_tree(self):
        index = list(UtilityLibrarian.existing_trees).index(self.target_tree)

        UtilityLibrarian.existing_trees[self.new_tree] = UtilityLibrarian.existing_trees.pop(self.target_tree)
        UtilityLibrarian.existing_output[self.new_tree] = UtilityLibrarian.existing_output.pop(self.target_tree)
        UtilityLibrarian.existing_sizes[self.new_tree] = UtilityLibrarian.existing_sizes.pop(self.target_tree)
        UtilityLibrarian.existing_symbols[self.new_tree] = UtilityLibrarian.existing_symbols.pop(self.target_tree)

        for index in list(UtilityLibrarian.existing_trees)[index:-1]:
            UtilityLibrarian.existing_trees[index] = UtilityLibrarian.existing_trees.pop(index)
            UtilityLibrarian.existing_output[index] = UtilityLibrarian.existing_output.pop(index)
            UtilityLibrarian.existing_sizes[index] = UtilityLibrarian.existing_sizes.pop(index)
            UtilityLibrarian.existing_symbols[index] = UtilityLibrarian.existing_symbols.pop(index)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class ResetSize:
    def __init__(self, tree):
        self.trees = tree

        self.validate_arguments()
        self.reset_size()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

    def reset_size(self):
        for tree in self.trees:
            UtilityLibrarian.existing_sizes.update({tree: 0})
            UtilityLibrarian.existing_symbols.update({tree: {"line": "┃", "split": "┣━", "end": "┗━"}})



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import os

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class ExportTree:
    def __init__(self, tree, output_folder, attributes=False):
        self.trees = tree
        self.output_folders = output_folder
        self.attributes = attributes

        self.validate_arguments()
        self.export_tree()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.output_folders, list):
            self.output_folders = [self.output_folders]

        for folder in self.output_folders:
            if not isinstance(folder, str):
                raise UtilityHandler.InstanceStringError("output_folder", folder)

            if not os.path.isdir(folder):
                raise UtilityHandler.NotExistingDirectoryError(folder)

        if self.attributes:
            if not isinstance(self.attributes, bool):
                raise UtilityHandler.InstanceBooleanError("attributes", self.attributes)

    def export_tree(self):
        for tree in self.trees:
            export_attributes = {tree: [UtilityLibrarian.existing_sizes[tree], UtilityLibrarian.existing_symbols[tree]]}

            for folder in self.output_folders:
                if not os.path.isdir(folder + "/br4nch-" + tree):
                    os.mkdir(folder + "/br4nch-" + tree)

                with open(
                        folder + "/br4nch-" + tree + "/tree-" + tree + ".br4nch", 'w', encoding='utf-8') as file:
                    file.write("tag=tree\n")
                    file.write(str({tree: UtilityLibrarian.existing_trees[tree]}))

                if self.attributes:
                    with open(
                            folder + "/br4nch-" + tree + "/attributes-" + tree + ".br4nch", 'w', encoding='utf-8') \
                            as file:
                        file.write("tag=attributes\n")
                        file.write(str(export_attributes))



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from os import path
from setuptools import setup, find_packages

from br4nch import __version__


with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as file:
    long_description = file.read()

setup(
    name="br4nch",
    version=__version__,
    description="Data Structure Tree Builder for Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TRSTN4/br4nch",
    author="TRSTN4",
    author_email="<tristan@trstn4.com>",
    classifiers=[
        "Intended Audience :: Developers",
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7"],
    keywords=["br4nch", "python", "python3", "data", "structure", "tree", "builder", "generate", "algorithm"],
    license="GNU General Public License v3.0",
    download_url="https://github.com/TRSTN4/br4nch/archive/refs/tags/" + __version__ + ".tar.gz",
    packages=find_packages(),
    install_requires=[],
    python_requires='>=3.7'
)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

import os
import ast

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class LoadTree:
    def __init__(self, tree_file, attributes_file=""):
        self.tree_file = tree_file
        self.attributes_file = attributes_file

        self.validate_arguments()
        self.load_tree()

    def validate_arguments(self):
        if not isinstance(self.tree_file, str):
            raise UtilityHandler.InstanceStringError("tree_file", self.tree_file)

        if not self.tree_file or self.tree_file and not os.path.isfile(self.tree_file):
            raise UtilityHandler.NotExistingTreeFileError(self.tree_file)

        with open(self.tree_file, 'r', encoding="utf8") as file:
            file = file.readlines()

            if str(file[0][:-1]) != "tag=tree":
                raise UtilityHandler.InvalidTreeFileError(self.tree_file)

            self.tree_file = ast.literal_eval(file[1])

        if list(self.tree_file)[0].lower() in list(map(str.lower, UtilityLibrarian.existing_trees)):
            raise UtilityHandler.DuplicateTreeError(list(self.tree_file)[0])

        if self.attributes_file:
            if not isinstance(self.attributes_file, str):
                raise UtilityHandler.InstanceStringError("attributes_file", self.attributes_file)

            if not self.attributes_file or self.attributes_file and not os.path.isfile(self.attributes_file):
                raise UtilityHandler.NotExistingAttributesFileError(self.attributes_file)

    def load_tree(self):
        if self.attributes_file:
            with open(self.attributes_file, 'r', encoding="utf8") as file:
                file = file.readlines()

                if str(file[0][:-1]) != "tag=attributes":
                    raise UtilityHandler.InvalidAttributesFileError(self.attributes_file)

                self.attributes_file = ast.literal_eval(file[1])

        UtilityLibrarian.existing_trees.update({list(self.tree_file)[0]: list(self.tree_file.values())[0]})
        UtilityLibrarian.existing_output.update({list(self.tree_file)[0]: []})

        if self.attributes_file and list(self.attributes_file)[0] == list(self.tree_file)[0] \
                and len(list(self.attributes_file.values())[0]) == 6:
            UtilityLibrarian.existing_sizes.update({list(self.tree_file)[0]: list(self.attributes_file.values())[0][0]})
            UtilityLibrarian.existing_symbols.update({list(
                self.tree_file)[0]: list(self.attributes_file.values())[0][1]})
        else:
            UtilityLibrarian.existing_sizes.update({list(self.tree_file)[0]: 0})
            UtilityLibrarian.existing_symbols.update({list(
                self.tree_file)[0]: {"line": "┃", "split": "┣━", "end": "┗━"}})



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler


class SetSymbol:
    def __init__(self, tree, line="", split="", end=""):
        self.trees = tree
        self.line = line
        self.split = split
        self.end = end

        self.validate_arguments()
        self.set_symbol()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if self.line:
            if not isinstance(self.line, str):
                raise UtilityHandler.InstanceStringError("line", self.line)

        if self.split:
            if not isinstance(self.split, str):
                raise UtilityHandler.InstanceStringError("split", self.split)

        if self.end:
            if not isinstance(self.end, str):
                raise UtilityHandler.InstanceStringError("end", self.end)

        if not self.line and not self.split and not self.end:
            raise UtilityHandler.RequiredSymbolChangeError

    def set_symbol(self):
        for tree in self.trees:
            if not self.line:
                self.line = UtilityLibrarian.existing_symbols[tree]["line"]
            if not self.split:
                self.split = UtilityLibrarian.existing_symbols[tree]["split"]
            if not self.end:
                self.end = UtilityLibrarian.existing_symbols[tree]["end"]

            UtilityLibrarian.existing_symbols[tree].update({"line": self.line, "split": self.split, "end": self.end})



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

__version__ = "2.2.1"
__description__ = "Data Structure Tree Builder"
__url__ = "https://github.com/TRSTN4/br4nch"

__author__ = "TRSTN4"
__author_email__ = "tristan@trstn4.com"
__author_website__ = "https://TRSTN4.com"

import os
import imp
import json
import urllib.request

from datetime import datetime, timedelta, date

from . import create, duplicate, move, replace, delete, set, reset, load, export, get, display, utility


class CheckVersion:
    def __init__(self):
        self.path = imp.find_module('br4nch')[1] + "/update_check.br4nch"

        self.check_version()

    def check_version(self):
        if not os.path.isfile(self.path):
            self.create_update_check_file()

        update_check_file = open(self.path, "rt")
        file = update_check_file.readlines()

        file_version = file[0]
        file_last_check_date = file[1]
        file_skip_check_hours = file[2]
        file_update_available = file[3]

        if file_version[8:-1] != __version__ and file_version[8:-1] != "None":
            file_version = file_version.replace(file_version[8:-1], "None")
            file_last_check_date = file_last_check_date.replace(file_last_check_date[16:-1], "None")
            file_skip_check_hours = file_skip_check_hours.replace(file_skip_check_hours[17:-1], "None")
            file_update_available = file_update_available.replace(file_update_available[17:], "None")

        if file_version[8:-1] == "None":
            file_version = file_version.replace(file_version[8:-1], __version__)

        if str(file_last_check_date[16:-1]) == "None" or datetime.now().strftime("%H") \
                not in file_skip_check_hours[17:-1].split(",") or datetime.now().strftime("%H") \
                in file_skip_check_hours[17:-1].split(",") \
                and date.today().strftime("%d") != file_last_check_date[16:-1]:
            file_last_check_date = file_last_check_date.replace(
                file_last_check_date[16:-1], date.today().strftime("%d"))

            skip_check = []
            for hour in range(4):
                skip_check.append((datetime.strptime(
                    datetime.now().strftime("%H"), "%H") + timedelta(hours=hour)).strftime("%H"))
            file_skip_check_hours = file_skip_check_hours.replace(file_skip_check_hours[17:-1], ",".join(skip_check))

            file_update_available = file_update_available.replace(file_update_available[17:], json.loads(
                urllib.request.urlopen(
                    "https://api.github.com/repos/TRSTN4/br4nch/releases/latest").read())["tag_name"])

        if file_update_available[17:] != __version__:
            print("[!] There is a new br4nch version (" + file_update_available[17:]
                  + ") available, The current installed br4nch " + "version (" + __version__ + ") is outdated.")

        update_check_file.close()

        update_check_file = open(self.path, "wt")
        update_check_file.write(file_version + file_last_check_date + file_skip_check_hours + file_update_available)
        update_check_file.close()

    def create_update_check_file(self):
        with open(self.path, 'w', encoding='utf-8') as file:
            file.write("version=None\n")
            file.write("last-check-date=None\n")
            file.write("skip-check-hours=None\n")
            file.write("update-available=None")


try:
    CheckVersion()
except:
    pass



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_generator import UtilityGenerator
from ..utility.utility_decider import UtilityDecider


class CreateNode:
    def __init__(self, tree, node, parent=""):
        self.trees = tree
        self.nodes = node
        self.parents = parent

        self.validate_arguments()
        self.build_position()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.nodes, list):
            self.nodes = [self.nodes]

        for index in range(len(self.nodes)):
            if not isinstance(self.nodes[index], str):
                raise UtilityHandler.InstanceStringError("node", self.nodes[index])

            while True:
                if self.nodes[index] and self.nodes[index][-1] == "\n":
                    self.nodes[index] = self.nodes[index][:-1]
                else:
                    break

        if not self.parents:
            self.parents = "0"

        if not isinstance(self.parents, list):
            self.parents = [self.parents]

    def build_position(self):
        for tree in self.trees:
            for position in UtilityDecider(tree, "parent", self.parents).get_formatted_positions():
                self.create_node(tree, position,
                                 UtilityLibrarian.existing_trees[tree][list(UtilityLibrarian.existing_trees[tree])[0]])

    def create_node(self, tree, position, nested_dictionary):
        if position[0] == "0":
            for node in self.nodes:
                UtilityLibrarian.existing_trees[tree][list(UtilityLibrarian.existing_trees[tree])[0]].update(
                    {node + UtilityGenerator().generate_uid(): {}})
            return

        count = 0
        for children in nested_dictionary.values():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    for node in self.nodes:
                        children.update({node + UtilityGenerator().generate_uid(): {}})
                    return
                else:
                    if children:
                        position.pop(0)
                        return self.create_node(tree, position, children)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from ..utility.utility_librarian import UtilityLibrarian
from ..utility.utility_handler import UtilityHandler
from ..utility.utility_decider import UtilityDecider


class DeleteNode:
    def __init__(self, tree, node):
        self.trees = tree
        self.nodes = node

        self.validate_arguments()
        self.manage_node()

    def validate_arguments(self):
        if not isinstance(self.trees, list):
            self.trees = [self.trees]

        if "*" in self.trees:
            self.trees.clear()
            for existing_tree in list(UtilityLibrarian.existing_trees):
                self.trees.append(existing_tree)

        for index in range(len(self.trees)):
            if not isinstance(self.trees[index], str):
                raise UtilityHandler.InstanceStringError("tree", self.trees[index])

            if self.trees[index].lower() not in list(map(str.lower, UtilityLibrarian.existing_trees)):
                raise UtilityHandler.NotExistingTreeError(self.trees[index])

            for existing_tree in list(UtilityLibrarian.existing_trees):
                if self.trees[index].lower() == existing_tree.lower():
                    self.trees[index] = existing_tree

        if not isinstance(self.nodes, list):
            self.nodes = [self.nodes]

    def manage_node(self):
        for tree in self.trees:
            queue_delete = []

            for position in UtilityDecider(tree, "node", self.nodes.copy()).get_formatted_positions():
                child = self.get_nodes(tree, position,
                                       UtilityLibrarian.existing_trees[tree][list(
                                           UtilityLibrarian.existing_trees[tree])[0]])
                queue_delete.append(child)

            for delete_child in queue_delete:
                if delete_child:
                    for parent_node, child_nodes in delete_child.items():
                        del child_nodes[parent_node]

    def get_nodes(self, tree, position, nested_dictionary):
        count = 0
        for parent, children in nested_dictionary.items():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    return {parent: nested_dictionary}
                else:
                    if children:
                        position.pop(0)
                        return self.get_nodes(tree, position, children)



# br4nch - Data Structure Tree Builder
# Author: https://TRSTN4.com
# Website: https://br4nch.com
# Documentation: https://docs.br4nch.com
# Github Repository: https://github.com/TRSTN4/br4nch

from br4nch.utility.utility_librarian import UtilityLibrarian
from br4nch.utility.utility_handler import UtilityHandler
from br4nch.utility.utility_generator import UtilityGenerator
from br4nch.display.display_tree import DisplayTree


class GetTree:
    def __init__(self, include="", exclude="", beautify=True):
        self.includes = include
        self.excludes = exclude
        self.beautify = beautify

        self.validate_arguments()
        self.log_tree()

    def validate_arguments(self):
        if self.includes:
            if not isinstance(self.includes, list):
                self.includes = [self.includes]

            for include in self.includes:
                if not isinstance(include, str):
                    raise UtilityHandler.InstanceStringError("include", include)

        if self.excludes:
            if not isinstance(self.excludes, list):
                self.excludes = [self.excludes]

            for exclude in self.excludes:
                if not isinstance(exclude, str):
                    raise UtilityHandler.InstanceStringError("exclude", exclude)

        if self.beautify:
            if not isinstance(self.beautify, bool):
                raise UtilityHandler.InstanceBooleanError("beautify", self.beautify)

    def log_tree(self):
        trees = []
        for tree in list(UtilityLibrarian.existing_trees):
            trees.append(tree)

        for tree in trees.copy():
            if self.includes:
                for include in self.includes:
                    if include not in tree:
                        trees.remove(tree)

            if self.excludes:
                for exclude in self.excludes:
                    if exclude in tree:
                        if tree in trees:
                            trees.remove(tree)

        if self.beautify:
            tree_uid = UtilityGenerator(True).generate_uid()

            UtilityLibrarian.existing_trees.update({tree_uid: {"Get Tree Result:": {}}})
            UtilityLibrarian.existing_output.update({tree_uid: []})
            UtilityLibrarian.existing_sizes.update({tree_uid: 0})
            UtilityLibrarian.existing_symbols.update({tree_uid: {"line": "┃", "split": "┣━", "end": "┗━"}})

            for tree in trees:
                UtilityLibrarian.existing_trees[tree_uid][list(UtilityLibrarian.existing_trees[tree_uid])[0]]\
                    .update({tree + UtilityGenerator().generate_uid(): {}})

            DisplayTree(tree_uid, True)
        else:
            for tree in trees:
                print(tree)
